--[[
Name: InventoryLib-2.0
Revision: $Rev: 34311 $
Author(s): Elkano (elkano@gmx.de)
Website: http://elkano.wowinterface.com/
Documentation: http://wiki.wowace.com/index.php/InventoryLib-2.0
SVN: http://svn.wowace.com/root/trunk/InventoryLib/InventoryLib-2.0
Description: A library that maintains a cache of linkable items and shares it between clients.
Dependencies: AceLibrary, AceComm-2.0, AceEvent-2.0, Gratuity-2.0
]]

local MAJOR_VERSION = "InventoryLib-2.0"
local MINOR_VERSION = "$Revision: 34311 $"

if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end
if not AceLibrary:HasInstance("AceEvent-2.0") then error(MAJOR_VERSION .. " requires AceEvent-2.0") end

local PATTERN_itemid = "item:(%-?%d+):%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+"
local PATTERN_itemlink = "|c%x+|H"..PATTERN_itemid.."|h%[.-%]|h|r"
local FORMAT_itemid = "item:%d:0:0:0:0:0:0:0"

local _G = getfenv(0)
local GetItemInfo = _G.GetItemInfo
local table_insert = _G.table.insert

local Gratuity
local InventoryLib = {}
-- -----
-- the following functions provide some tools
-- -----
local function getbaseid(link)
	if link then
		if type(link) == "number" then return link end
		local _, _, baseid = link:find(PATTERN_itemlink)
		if not baseid then
			_, _, baseid = link:find(PATTERN_itemid)
		end
		if not baseid then
			baseid = link
		end
		return tonumber(baseid)
	end
	return
end

-- -----
-- public functions
-- -----

-- returns a table of all itemIDs in cache
-- the list is rescanned step by step in background or if forcedRescan is set
function InventoryLib:GetItems(forceRescan)
	if forceRescan then
		self:BuildLists(true)
	end
	return self.localvars.items
end

-- returns true if the item from link is in the client's cache
function InventoryLib:IsCacheItem(item)
	item = getbaseid(item)
	self:BuildLists(false, item)
	return (self.localvars.items[item] ~= nil)
end

-- returns true if the item from link is known as valid for this session
function InventoryLib:IsSessionItem(item)
	item = getbaseid(item)
	self:BuildLists(false, item)
	return (self.localvars.items[item] == true)
end

function InventoryLib:GetTypesTables()
	return self.localvars.types_tree, self.localvars.types_base, self.localvars.types_sub, self.localvars.types_slot
end

function InventoryLib:GetCounts()
	return self.localvars.count_known, self.localvars.count_safe
end
-- -----
-- 'internal' functions
-- -----

-- this function scans the client's cache for items
-- it will also build a tree for the various types an item can have and sortet lists for the types and subtypes
function InventoryLib:BuildLists(forceRescan, scanid)
	self:RecheckSessionItems()

	local localvars = self.localvars

	local newtype = false
	local first, last
	if forceRescan == true then
		first = 1
		last = self.consts.maxid
		localvars.items_scan_last = 0
		scanid = nil
	else
		if scanid then
			first = scanid
			last = scanid
		else
			first = localvars.items_scan_last + 1
			last = localvars.items_scan_last + self.consts.itemscancount
			localvars.items_scan_last = (last < self.consts.maxid) and last or 0
		end
	end
	local types_tree = localvars.types_tree
	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc
	for i = first, last do
		if localvars.items[i] == nil or scanid then
			itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc = GetItemInfo(i)
			if itemName then	-- valid item
				itemEquipLoc = _G[itemEquipLoc]
				if (not itemEquipLoc or itemEquipLoc == "") then
					itemEquipLoc = self.locals.INVTYPE_OTHER
				end
				if localvars.items[i] == nil then
					localvars.items[i] = false
					localvars.count_known = localvars.count_known + 1
				end
				if (not types_tree[itemType]) then
					types_tree[itemType] = {}
					newtype = true
				end
				if (not types_tree[itemType][itemSubType]) then
					types_tree[itemType][itemSubType] = {}
					newtype = true
				end
				if (not types_tree[itemType][itemSubType][itemEquipLoc]) then
					types_tree[itemType][itemSubType][itemEquipLoc] = ""
					newtype = true
				end
			end
		end
	end
	if not newtype then return end
	for k in pairs(localvars.types_base) do
		localvars.types_base[k] = nil
	end
	for k in pairs(localvars.types_base) do
		localvars.types_base[k] = nil
	end
	for k in pairs(localvars.types_sub) do
		localvars.types_sub[k] = nil
	end
	for k in pairs(localvars.types_slot) do
		localvars.types_slot[k] = nil
	end
	for type, subtypes in pairs(localvars.types_tree) do
		table_insert(localvars.types_base, type)
		local subtypesort = {}
		local hasinvtypes = false
		for subtype, invtypes in pairs(subtypes) do
			table_insert(subtypesort, subtype)
			local invtypesort = {}
			for invtype, _ in pairs(invtypes) do
				table_insert(invtypesort, invtype)
			end
			if #invtypesort > 1 then
				table.sort(invtypesort)
				if (not localvars.types_slot[type]) then
					localvars.types_slot[type] = {}
				end
				localvars.types_slot[type][subtype] = invtypesort
				hasinvtypes = true
			end
		end
		if #subtypesort > 1 or hasinvtypes then
			table.sort(subtypesort)
			localvars.types_sub[type] = subtypesort
		end
	end
	table.sort(localvars.types_base)
end

function InventoryLib:StoreSessionItem(item, viacomm)

	local localvars = self.localvars

	item = getbaseid(item)
	if not item then return false end
	if (not localvars.items[item]) then
		if localvars.items[item] == nil then	-- item not mined from cache, yet
			local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc = GetItemInfo(item)
			if not itemName then	-- item not cached, yet
				Gratuity:SetHyperlink(FORMAT_itemid:format(item))
				localvars.items_rescan[item] = not viacomm and true or false
				return false
			end
			itemEquipLoc = _G[itemEquipLoc]
			if (not itemEquipLoc or itemEquipLoc == "") then
				itemEquipLoc = self.locals.INVTYPE_OTHER
			end
			localvars.items[item] = false
			localvars.count_known = localvars.count_known + 1
			local types_tree = localvars.types_tree
			if not (types_tree and types_tree[itemType] and types_tree[itemType][itemSubType] and types_tree[itemType][itemSubType][itemEquipLoc]) then -- at least one of the type strings isn't known yet so rebuild type trees
				self:BuildLists(false, item)
			end
		end
		localvars.items[item] = true
		localvars.count_safe = localvars.count_safe + 1
		if not viacomm then	-- ignore items recived via comm for priority sharing
			table_insert(localvars.items_new, item)
		end
	end
	return true
end

function InventoryLib:RecheckSessionItems()
	for item, viacomm in pairs(self.localvars.items_rescan) do
		if InventoryLib:StoreSessionItem(item, viacomm) then
			self.localvars.items_rescan[item] = nil
		end
	end
end

-- -----
-- the following functions scan various sources for items in order to add them to the list of linkable items for this session
-- -----

-- if given unit hasn't been inspected lately all slots will be scaned for items
function InventoryLib:Inspect(unit)
	if not (UnitExists(unit) and UnitIsPlayer(unit) and UnitIsFriend(unit, "player") and CheckInteractDistance(unit, 1)) then
		return
	end
	local timenow = time()
	if timenow - self.localvars.inspected_lastclean > 5 then -- clean up cache at most every 5 seconds
		for key, value in pairs(self.localvars.inspected_units) do
			if (timenow - value > 300) then -- remove after 5min
				self.localvars.inspected_units[key] = nil
			end
		end
		self.localvars.inspected_lastclean = timenow
	end
	local name, server = UnitName(unit)
	if not server then server = self.consts.RealmName end
	if self.localvars.inspected_units[name.."-"..server] then
		return
	end
	self.localvars.inspected_units[name.."-"..server] = timenow
	for _, slot in ipairs(self.consts.InventorySlotName) do
		local item = GetInventoryItemLink(unit, GetInventorySlotInfo(slot))
		if item then
			self:StoreSessionItem(item)
		end
	end
end

function InventoryLib:ScanAuctionPage()
	local numBatchAuctions = GetNumAuctionItems("list")
	for auctionid = 1, numBatchAuctions do
		local item = GetAuctionItemLink("list", auctionid)
		if item then
			self:StoreSessionItem(item)
		end
	end
end

function InventoryLib:ScanBags()
	for bagid = 0, NUM_BAG_SLOTS do
		local size = GetContainerNumSlots(bagid)
		if size then
			for slotid = 1, size do
				local item = GetContainerItemLink(bagid, slotid)
				if item then
					self:StoreSessionItem(item)
				end
			end
		end
	end
end

function InventoryLib:ScanBank()
	for bagid_i = NUM_BAG_SLOTS, (NUM_BAG_SLOTS + NUM_BANKBAGSLOTS) do
		local bagid = bagid_i
		if bagid == NUM_BAG_SLOTS then
			bagid = BANK_CONTAINER
		end
		local size = GetContainerNumSlots(bagid)
		if size then
			for slotid = 1, size do
				local item = GetContainerItemLink(bagid, slotid)
				if item then
					self:StoreSessionItem(item)
				end
			end
		end
	end
end

function InventoryLib:ScanMerchant()
	if MerchantFrame:IsVisible() and MerchantFrame.selectedTab == 1 then
		local numMerchantItems = GetMerchantNumItems()
		for i = 1, numMerchantItems do
			local item = GetMerchantItemLink(i)
			if item then
				self:StoreSessionItem(item)
			end
		end
	end
end

function InventoryLib:ScanTradeSkill()
	if TradeSkillFrame ~= nil and TradeSkillFrame:IsVisible() then
		for i=1, GetNumTradeSkills() do
			local _, type, _, _ = GetTradeSkillInfo(i)
			if type ~= "header" then
				local item = GetTradeSkillItemLink(i)
				if item then
					self:StoreSessionItem(item)
				end
				for j=1, GetTradeSkillNumReagents(i) do
					local reagentName, _, _, _ = GetTradeSkillReagentInfo(i, j)
					if reagentName then
						local item = GetTradeSkillReagentItemLink(i, j)
						if item then
							self:StoreSessionItem(item)
						end
					end
				end
			end
		end
	end
end

function InventoryLib:ScanCraft()
	if CraftFrame ~= nil and CraftFrame:IsVisible() then
		for i=1, GetNumCrafts() do
			local _, type, _, _ = GetCraftInfo(i)
			if type ~= "header" then
				local item = GetCraftItemLink(i)
				if item then
					self:StoreSessionItem(item)
				end
				for j=1, GetCraftNumReagents(i) do
					local reagentName, _, _, _ = GetCraftReagentInfo(i, j)
					if reagentName then
						local item = GetCraftReagentItemLink(i, j)
						if item then
							self:StoreSessionItem(item)
						end
					end
				end
			end
		end
	end
end

-- -----
-- lib stuff
-- -----

local function activate(self, oldLib, oldDeactivate)
	InventoryLib = self
	if oldLib then
		self.localvars = oldLib.localvars

		oldlib:CancelAllScheduledEvents()
		oldlib:UnregisterAllEvents()

		if oldLib.commPrefix then
			oldLib.UnregisterAllComms(self) -- AceLibrary's magic reference-changing
			self.commPrefix = oldLib.commPrefix
		end
	else
		self.localvars = {
			["framescan_last"]			= 0,			-- timestamp of latest frame scan (only one per second done)
			["inspected_lastclean"]		= 0,			-- timestamp of the last pruning of scanned units
			["inspected_units"]			= {},			-- list of latest scanned units
			["items"]					= {},			-- list of all known items keyed by id, true if session valid
			["items_scan_last"]			= 0,			-- id of the last scanned item; will scan consts.itemscancount items every second
			["items_new"]				= {},			-- list of new items seen this session, used for propagation
			["items_propagated"]		= 0,			-- id of the last nonnew item propagated
			["items_rescan"]			= {},			-- list of valid items not found in cache, yet
			["types_tree"]				= {},
			["types_base"]				= {},
			["types_sub"]				= {},
			["types_slot"]				= {},
			["count_known"]				= 0,
			["count_safe"]				= 0,
			["temp_table"]				= {},			-- table to be used inside functions; clear entries after use
		}
	end

	self.consts = {
		-- list of all slots scanned on inspect
		["InventorySlotName"] = {
			"HeadSlot",
			"NeckSlot",
			"ShoulderSlot",
			"BackSlot",
			"ChestSlot",
			"ShirtSlot",
			"TabardSlot",
			"WristSlot",
			"HandsSlot",
			"WaistSlot",
			"LegsSlot",
			"FeetSlot",
			"Finger0Slot",
			"Finger1Slot",
			"Trinket0Slot",
			"Trinket1Slot",
			"MainHandSlot",
			"SecondaryHandSlot",
			"RangedSlot",
			"AmmoSlot",
			"Bag0Slot",
			"Bag1Slot",
			"Bag2Slot",
			"Bag3Slot",
		},
		["itemscancount"] = 100,
		["maxid"] = 40000,
		["itemscancount_comm"] = 1000,
		["itemsendcount_comm"] = 5,
		["RealmName"] = GetRealmName()
	}

	self.locals = {
		["INVTYPE_OTHER"] = "Other",
	}

	if ( GetLocale() == "deDE" ) then
		self.locals.INVTYPE_OTHER = "Sonstiges"
	end

	if oldDeactivate then
		oldDeactivate(oldLib)
	end
end

local function external(self, major, instance)
	if major == "AceEvent-2.0" then
		AceEvent = instance
		AceEvent:embed(self)
		self:UnregisterAllEvents()
		self:CancelAllScheduledEvents()
		if AceEvent:IsFullyInitialized() then
			self:AceEvent_FullyInitialized()
		else
			self:RegisterEvent("AceEvent_FullyInitialized", "AceEvent_FullyInitialized", true)
		end		
	elseif major == "AceComm-2.0" then
		instance:embed(self)
		if not self.commPrefix then
			self:SetCommPrefix("InvL")
		end
		self:RegisterComm(self.commPrefix, "GROUP")
		self:RegisterComm(self.commPrefix, "GUILD")
		self:RegisterComm(self.commPrefix, "WHISPER")
--[[	elseif major == "AceDebug-2.0" then
		instance:embed(self)
		self.debugFrame = ChatFrame4
		self:SetDebugging(true) ]]
	elseif major == "Gratuity-2.0" then
		Gratuity = instance
	end
end

-- -----
-- events
-- -----
local ChatMessageTypes = {
	"CHAT_MSG_SYSTEM",
	"CHAT_MSG_SAY",
	"CHAT_MSG_TEXT_EMOTE",
	"CHAT_MSG_YELL",
	"CHAT_MSG_WHISPER",
	"CHAT_MSG_WHISPER_INFORM",
	"CHAT_MSG_PARTY",
	"CHAT_MSG_GUILD",
	"CHAT_MSG_OFFICER",
	"CHAT_MSG_CHANNEL",
	"CHAT_MSG_RAID",
	"CHAT_MSG",
	"CHAT_MSG_LOOT",
	"CHAT_MSG_TRADE",
}

function InventoryLib:AceEvent_FullyInitialized()
	self:RegisterEvent("AUCTION_ITEM_LIST_UPDATE", "Event_Auction")
	self:RegisterEvent("BANKFRAME_OPENED", "Event_Bank")
	self:RegisterEvent("MERCHANT_SHOW", "Event_Merchant")
	self:RegisterBucketEvent("MERCHANT_UPDATE", 1, "Event_Merchant")
	self:RegisterEvent("TRADE_SKILL_SHOW", "Event_Tradeskill")
	self:RegisterBucketEvent("TRADE_SKILL_UPDATE", 1, "Event_Tradeskill")
	self:RegisterEvent("CRAFT_SHOW", "Event_Craft")
	self:RegisterBucketEvent("CRAFT_UPDATE", 1, "Event_Craft")

	self:ScheduleRepeatingEvent(self.BuildLists, 1, self)
	self:ScheduleRepeatingEvent(self.ShareItems, 1, self)

	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")

	for _, ChatMessageType in pairs(ChatMessageTypes) do
		self:RegisterEvent(ChatMessageType, "Event_Chat")
	end

	self:BuildLists(true)
	self:Inspect("player")
	self:ScanBags()
end

function InventoryLib:Event_Auction()
	self:ScanAuctionPage()
end

function InventoryLib:Event_Bank()
	self:Inspect("player")
	self:ScanBags()
	self:ScanBank()
end

function InventoryLib:Event_Merchant()
	self:ScanMerchant()
end

function InventoryLib:Event_Tradeskill()
	self:ScanTradeSkill()
end

function InventoryLib:Event_Craft()
	self:ScanCraft()
end

function InventoryLib:PLAYER_ENTERING_WORLD()
end

function InventoryLib:PLAYER_TARGET_CHANGED()
	if UnitExists("target") and UnitIsPlayer("target") then
		self:Inspect("target")
	end
end

function InventoryLib:UPDATE_MOUSEOVER_UNIT()
	if UnitExists("mouseover") and UnitIsPlayer("mouseover") then
		self:Inspect("mouseover")
	end
end

function InventoryLib:Event_Chat(msg)
	for id in msg:gmatch(PATTERN_itemlink) do
		self:StoreSessionItem(id)
	end
end

------------------------------------------------
-- AceComm
------------------------------------------------

local COMM_COMPAT = 1
InventoryLib.commHandlers = {}

function InventoryLib:OnCommReceive(prefix, sender, distribution, compat, handler, ...)
	if compat < COMM_COMPAT then
		self:SendCommMessage("WHISPER", sender, COMM_COMPAT, "OLDVERSION", MAJOR_VERSION, MINOR_VERSION)
	elseif compat > COMM_COMPAT then
		self.commHandlers:OLDVERSION(sender, "unknown", "")
	elseif type(handler) == 'string' and self.commHandlers[handler] then
		self.commHandlers[handler](self, sender, ...)
	end
end

function InventoryLib.commHandlers:OLDVERSION(sender, major, minor)
	print(string.format("You have an older version of the InventoryLib library (%s - %s) than %s (%s - %s) -- please upgrade the addon which is using it.", MAJOR_VERSION, MINOR_VERSION, sender, major, minor))
end

function InventoryLib.commHandlers:VI(sender, ...) -- validate items
	for i = 1, select('#', ...) do
		self:StoreSessionItem(select(i, ...), true)
	end
end

function InventoryLib:ShareItems()
	if not self.SendCommMessage then return end
	local consts = self.consts
	local localvars = self.localvars
	local items = localvars.temp_table
	local items_new = localvars.items_new
	local itemscancount_comm = consts.itemscancount_comm
	local itemsendcount_comm = consts.itemsendcount_comm
	while #items_new > 0 and #items < itemsendcount_comm do
		table_insert(items, table.remove(items_new, math.random(#items_new)))
	end
	if #items < itemsendcount_comm then
		local i = 0
		local item_propagate = localvars.items_propagated
		while #items < itemsendcount_comm and i < itemscancount_comm do
			item_propagate = (item_propagate < consts.maxid) and (item_propagate + 1) or 1
			if localvars.items[item_propagate] == true then
				table_insert(items, item_propagate)
			end
			i = i + 1
		end
		localvars.items_propagated = item_propagate
	end
	if #items > 0 then
		self:SendCommMessage("GROUP", COMM_COMPAT, "VI", unpack(items))
		self:SendCommMessage("GUILD", COMM_COMPAT, "VI", unpack(items))
		for k in pairs(items) do items[k] = nil end
	end
end

-- -----
-- load the lib
-- -----
AceLibrary:Register(InventoryLib, MAJOR_VERSION, MINOR_VERSION, activate, nil, external)
